library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.emp_data_types.all;
use work.dtc_link_maps.all;
use work.dtc_constants.all;
use work.dtc_data_types.all;


entity LinkCombinerCore is
    generic (
        N_INPUT_LINKS  : integer := 8
    );
    port (
        --- Input Ports ---
        clk_p            : in std_logic;
        reset            : in std_logic;
        output_reset     : in std_logic := '0';
        output_enable    : in std_logic := '0';
        links_in         : in ldata(N_INPUT_LINKS - 1 downto 0);
        packet_start     : in std_logic;
        header_in        : in tCICHeaderArray(cNumberOfFEModules * cNumberOfCICs - 1 downto 0) := (others => ('0', (others => '0'), (others => '0'), (others => '0')));
        header_user_bits : in std_logic_vector(31 downto 0) := (others => '0');
        --- Ouput Ports ---
        link_out        : out lword := LWORD_NULL;
        --- Debug Ports ---
        debug            : out lword := LWORD_NULL;
        debug_super_id   : out std_logic_vector(31 downto 0);
        debug_fifo       : out std_logic_vector(31 downto 0) := (others => '0')
    );
end LinkCombinerCore;

architecture Behavioral of LinkCombinerCore is

    constant cWordWidth       : integer := 64;
    constant cOutputWordWidth : integer := 64;

    COMPONENT link_combin_buffer
      PORT (
        clk          : IN STD_LOGIC;
        srst         : IN STD_LOGIC;
        din          : IN STD_LOGIC_VECTOR(cOutputWordWidth - 1 DOWNTO 0);
        wr_en        : IN STD_LOGIC;
        rd_en        : IN STD_LOGIC;
        dout         : OUT STD_LOGIC_VECTOR(cOutputWordWidth - 1 DOWNTO 0);
        full         : OUT STD_LOGIC;
        almost_full  : OUT STD_LOGIC;
        overflow     : OUT STD_LOGIC;
        empty        : OUT STD_LOGIC;
        almost_empty : OUT STD_LOGIC;
        valid        : OUT STD_LOGIC;
        underflow    : OUT STD_LOGIC;
        data_count   : OUT STD_LOGIC_VECTOR(7 DOWNTO 0);
        wr_rst_busy  : OUT STD_LOGIC;
        rd_rst_busy  : OUT STD_LOGIC
      );
    END COMPONENT;
    
    COMPONENT link_agg_all_links_fifo
    PORT (
        --- Input Ports ---
        clk          : IN STD_LOGIC;
        srst         : IN STD_LOGIC;
        din          : IN STD_LOGIC_VECTOR(cWordWidth - 1 DOWNTO 0);
        wr_en        : IN STD_LOGIC;
        rd_en        : IN STD_LOGIC;
        --- Output Ports ---
        dout         : OUT STD_LOGIC_VECTOR(cWordWidth - 1 DOWNTO 0);
        full         : OUT STD_LOGIC;
        overflow     : OUT STD_LOGIC;
        empty        : OUT STD_LOGIC;
        almost_empty : OUT STD_LOGIC;
        valid        : OUT STD_LOGIC;
        data_count   : OUT STD_LOGIC_VECTOR(5 DOWNTO 0);
        wr_rst_busy  : OUT STD_LOGIC;
        rd_rst_busy  : OUT STD_LOGIC
    );
    END COMPONENT;

    -- Type Definitions
    type tWordArray            is array(integer range <>) of std_logic_vector(cWordWidth - 1 downto 0);
    type tStatusShiftRegister  is array(1 downto 0)       of std_logic_vector(cNumberOfFEModules * cNumberOfCICs * 9 - 1 downto 0);
    type tBCIDShiftRegister    is array(1 downto 0)       of std_logic_vector(12 - 1 downto 0);
    type tInputDataCountArray  is array(integer range <>) of std_logic_vector(5 downto 0);

    -- Constants
    constant cNullValidArray                                              : std_logic_vector(N_INPUT_LINKS - 1 downto 0)       := (others => '0');
    constant cFullEmptyArray                                              : std_logic_vector(N_INPUT_LINKS - 1 downto 0)       := (others => '1');

    -- Signals    
    signal input_dout_array                                               : tWordArray(0 to N_INPUT_LINKS - 1)                 := (others => (others => '0'));
    signal input_rd_en_array                                              : std_logic_vector(N_INPUT_LINKS - 1 downto 0)       := (others => '0');
    signal input_almost_empty_array, input_valid_array                    : std_logic_vector(N_INPUT_LINKS - 1 downto 0)       := (others => '0');
    signal input_empty_array                                              : std_logic_vector(N_INPUT_LINKS - 1 downto 0)       := (others => '0');
    signal input_data_valid                                               : std_logic                                          := '0';
    signal input_data_count                                               : tInputDataCountArray(N_INPUT_LINKS - 1 downto 0)   := (others => (others => '0'));
    
    signal output_din, output_din_buf                         : std_logic_vector(cOutputWordWidth - 1 downto 0)          := (others => '0');
    signal output_wr_en, output_wr_en_buf, output_full  : std_logic      := '0';
    signal output_almost_full, output_overflow                : std_logic      := '0';
    signal output_pointer                                                 : integer range 0 to N_INPUT_LINKS - 1               := 0;
    signal output_pointer_buffered                                        : integer range 0 to N_INPUT_LINKS - 1               := 0;
    signal output_data_count                                              : std_logic_vector(7 downto 0) := (others => '0');
    signal output_empty                          : std_logic;
    signal output_valid                          : std_logic;
    signal output_dout                           : std_logic_vector(cOutputWordWidth - 1 downto 0);
    signal output_almost_empty                   : std_logic;
    signal output_underflow                      : std_logic;
    signal output_readout_countdown              : integer range 0 to 256       := 0;
    signal output_valid_previous                 : std_logic                    := '0';
    signal output_strobe_buffer, output_strobe_buffer_2 : std_logic                    := '0';
    signal output_rd_en_check                    : std_logic                    := '0';
    signal output_rd_en                    : std_logic                    := '0';
    
    signal counter                                                        : integer                                            := 0;
    signal status_sr                                                      : tStatusShiftRegister                               := (others => (others => '0'));
    signal bcid_sr                                                        : tBCIDShiftRegister                                 := (others => (others => '0'));
    signal super_id                                                       : unsigned(32 - 1 downto 0)                          := (others => '0');

    signal packet_stub_count, packet_stub_count_previous                  : unsigned(7 downto 0)                               := (others => '0');
    signal packet_stub_counter_pointer                                    : integer                                            := 0;
    signal packet_stub_count_done                                         : std_logic                                          := '0';


begin

    --==============================--
    genInputLinkBuffers : for i in 0 to N_INPUT_LINKS - 1 generate
    --==============================--
        signal full  : std_logic;
        signal din   : std_logic_vector(cWordWidth - 1 downto 0) := (others => '0');
        signal wr_en : std_logic                                 := '0';
    begin
    
        --==============================--
        InputLinkBufferInstance : link_agg_all_links_fifo
        --==============================--
        Port Map (
            clk          => clk_p,
            srst         => reset,
            din          => din,
            wr_en        => wr_en,
            rd_en        => input_rd_en_array(i),
            dout         => input_dout_array(i),
            full         => full,
            empty        => input_empty_array(i),
            almost_empty => input_almost_empty_array(i),
            valid        => input_valid_array(i),
            data_count   => input_data_count(i)
        );
        
        --==============================--
        pBufferInputs : process(clk_p)
        --==============================--
        begin
            if rising_edge(clk_p) then
                if links_in(i).valid = '1' then
                    din   <= links_in(i).data;
                    wr_en <= '1';
                else
                    din   <= (others => '0');
                    wr_en <= '0';
                end if;
            end if;
        end process pBufferInputs;
        
    end generate genInputLinkBuffers;

    
    --==============================--
    OutputLinkBufferInstance : link_combin_buffer
    --==============================--
    PORT MAP (
        clk          => clk_p,
        srst         => output_reset,
        din          => output_din_buf,
        wr_en        => output_wr_en_buf,
        rd_en        => output_rd_en,
        dout         => output_dout,
        empty        => output_empty,
        almost_full  => output_almost_full,
        full         => output_full,
        overflow     => output_overflow,
        valid        => output_valid,
        underflow    => output_underflow,
        data_count   => output_data_count,
        almost_empty => output_almost_empty
    );
    
    --==============================--
    pOutput: process(clk_p)
    --==============================--
        variable data_count : unsigned(7 downto 0);
    begin
        if rising_edge(clk_p) then
            output_valid_previous <= output_valid;
            if output_valid = '1' and output_rd_en_check = '1' then
                link_out.valid <= '1';
                if output_valid_previous = '0' then
                    link_out.start <= '1';
                else
                    link_out.start <= '0';
                end if;
                link_out.strobe <= output_strobe_buffer_2;
                link_out.data   <= output_dout;
            else
                link_out <= LWORD_NULL;
            end if;
            
            data_count := unsigned(output_data_count);
            if input_data_valid = '0' then
                if data_count >= 2 and output_readout_countdown = 0 then
                    output_rd_en_check       <= output_enable;
                    output_readout_countdown <= to_integer(data_count);
                elsif output_almost_full = '1' and output_readout_countdown = 0 then
                    output_rd_en_check       <= output_enable;
                    output_readout_countdown <= to_integer(data_count);
                end if;
            else
                if output_almost_full = '1' and output_readout_countdown = 0 then
                    output_rd_en_check <= output_enable;
                    output_readout_countdown <= to_integer(data_count);
                end if;
            end if;
            if output_readout_countdown > 0 then
                output_readout_countdown <= output_readout_countdown - 1;
                output_rd_en <= '1';
                if output_readout_countdown = 1 then
                    output_strobe_buffer <= '1';
                else
                    output_strobe_buffer <= '0';
                end if;
            else
                output_rd_en <= '0';
                output_strobe_buffer <= '0';
            end if;
            
            output_strobe_buffer_2 <= output_strobe_buffer;
        end if;
    end process pOutput;
    
    
    
    --==============================--
    pValidDataChecker : process(clk_p)
    --==============================--
    begin
        if rising_edge(clk_p) then
            if packet_start = '1' then
                input_data_valid <= '1';
            else
                if reset = '1' then
                    input_data_valid <= '0';
                end if;
            end if;
        end if;
    end process pValidDataChecker;
    
    
    
    --==============================--
    pCombineBuffers : process(clk_p)
    --==============================--
        variable next_pointer_location : integer := 0;
        variable valid_array_empty : std_logic := '0';
    begin
        if rising_edge(clk_p) then
            -- Iterate pointer to successively empty the input FIFOs
            if input_data_valid = '1' and counter = 4 then
                if input_almost_empty_array(output_pointer) = '0' then
                    input_rd_en_array(output_pointer) <= '1';
                else
                    if output_pointer = N_INPUT_LINKS - 1 then
                        next_pointer_location := 0;
                    else
                        next_pointer_location := next_pointer_location + 1;
                    end if;
                    input_rd_en_array(output_pointer) <= '0';
                    output_pointer <= next_pointer_location;
                    if input_empty_array(next_pointer_location) = '0' then
                        input_rd_en_array(next_pointer_location) <= '1';
                    else
                        input_rd_en_array(next_pointer_location) <= '0';
                    end if;
                end if;
            else
                input_rd_en_array(output_pointer) <= '0';
            end if;
            
            -- Packet control bits logic
            if packet_start = '1' then
                counter <= 0;
            else
                if counter < 4 then
                    counter <= counter + 1;
                else
                    counter <= counter;
                end if;
            end if;
            output_pointer_buffered <= output_pointer;
            output_din     <= input_dout_array(output_pointer_buffered);
            output_wr_en   <= input_valid_array(output_pointer_buffered);
            
            
            if reset = '1' then
                packet_stub_count <= (others => '0');
                packet_stub_count_previous <= packet_stub_count;
                packet_stub_counter_pointer <= 0;
                packet_stub_count_done <= '0';
            else
                if packet_stub_count_done = '0' then
                    if packet_stub_counter_pointer = (cNumberOfFEModules - 1) * cNumberOfCICs then
                        packet_stub_count_done <= '1';
                    end if;
                    packet_stub_count <= packet_stub_count + unsigned(header_in(packet_stub_counter_pointer).stub_count) + unsigned(header_in(packet_stub_counter_pointer + 1).stub_count);
                    packet_stub_counter_pointer <= packet_stub_counter_pointer + cNumberOfCICs;
                end if;
            end if;
            
            if counter = 1 then
                output_wr_en_buf            <= '1';
                output_din_buf(63 downto 32) <= header_user_bits;
                output_din_buf(31 downto 0)  <= std_logic_vector(super_id);
            elsif counter = 2 then
                output_din_buf <= (others => '0');
                output_din_buf(63 downto 56) <= std_logic_vector(packet_stub_count_previous);
                output_din_buf(47 downto 36) <= bcid_sr(bcid_sr'high);
                output_wr_en_buf <= '1';
                if cNumberOfFEModules < 3 then
                    output_din_buf(cNumberOfFEModules * cNumberOfCICs * 9 - 1 downto 0) <= status_sr(status_sr'high);
                else
                    output_din_buf(2 * cNumberOfCICs * 9 - 1 downto 0) <= status_sr(status_sr'high)(2 * cNumberOfCICs * 9 - 1 downto 0);
                end if;
            elsif counter = 3 then
                output_din_buf <= (others => '0');
                output_wr_en_buf <= '1';
                if cNumberOfFEModules >= 3 and cNumberOfFEModules < 6 then
                    output_din_buf((cNumberOfFEModules - 2) * cNumberOfCICs * 9 - 1 downto 0) <= status_sr(status_sr'high)(cNumberOfFEModules * cNumberOfCICs * 9 - 1 downto 2 * cNumberOfCICs * 9);
                end if;
            else
                output_wr_en_buf  <= output_wr_en;
                output_din_buf <= output_din;
            end if;
            
        end if;
    end process pCombineBuffers;


    --==============================--
    pBufferHeader: process(clk_p)
    --==============================--
        variable status           : std_logic_vector(cNumberOfFEModules * cNumberOfCICs * 9 - 1 downto 0) := (others => '0');
        variable bcid             : std_logic_vector(12 - 1 downto 0)                                     := (others => '0');
    begin
        if rising_edge(clk_p) then
            if output_reset = '1' then
                super_id <= (others => '0');
            else
                if packet_start = '1' then
                    for i in cNumberOfFEModules - 1 downto 0 loop
                        status(i*cNumberOfCICs*9 + 8  downto i*cNumberOfCICs*9)     := std_logic_vector(header_in(i*cNumberOfCICs).status);
                        status(i*cNumberOfCICs*9 + 17 downto i*cNumberOfCICs*9 + 9) := std_logic_vector(header_in(i*cNumberOfCICs + 1).status);
                    end loop;
                    status_sr <= status_sr(status_sr'high - 1 downto 0) & status;
                    bcid      := std_logic_vector(header_in(0).bcid);
                    bcid_sr   <= bcid_sr(bcid_sr'high - 1 downto 0) & bcid;
                    if unsigned(bcid_sr(bcid_sr'high)) > unsigned(bcid_sr(bcid_sr'high - 1)) then
                        super_id <= super_id + 1;
                    end if;
                end if;
            end if;
        end if;
    end process pBufferHeader;

    --==============================--
    -- Debug
    --==============================--
    debug.data(31 downto 0) <= std_logic_vector(super_id);
    debug.data(39 downto 32) <= output_data_count;
    debug.data(47 downto 40) <= "00" & input_data_count(0);
    debug.data(51 downto 48) <= std_logic_vector(to_unsigned(counter, 4));
    debug.data(52) <= output_rd_en;
    debug.data(56) <= packet_start;
    debug.data(57) <= reset;
    debug.data(58) <= input_data_valid;
    debug.valid <= '1';
    debug.strobe <= '1';
    debug_super_id <= std_logic_vector(super_id);
    debug_fifo(7 downto 0) <= output_data_count;
    debug_fifo(8) <= output_full;
    debug_fifo(9) <= output_rd_en;
    debug_fifo(10) <= output_overflow;

end Behavioral;
