library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.emp_data_types.all;

use work.ipbus.all;
use work.ipbus_reg_types.all;

use work.ipbus_decode_mprocessor.all;

use work.dtc_link_maps.all;
use work.mprocessor_constants.all;
use work.dtc_constants.all;
use work.dtc_data_types.all;


entity MProcessor is
    generic (
        N_LINKS : integer
    );
    port (
        --- Input Ports ---
        clk_p              : in  std_logic;
        links_in           : in  ldata(N_LINKS - 1 downto 0);
        header_in          : in  tCICHeaderArray(cNumberOfFEModules * cNumberOfCICs - 1 downto 0) := (others => ('0', (others => '0'), (others => '0'), (others => '0')));
        header_start_array : in  tHeaderStartArray                                                := (others => (others => '0'));
        gbe_backpressure   : in  std_logic                                                        := '0';
        srst               : in  std_logic                                                        := '0';
        --- Output Ports ---
        link_out           : out lword                                                            := LWORD_NULL;
        --- IPBus Ports ---
        ipb_clk            : in  std_logic;
        ipb_rst            : in  std_logic;
        ipb_in             : in  ipb_wbus;
        ipb_out            : out ipb_rbus
    );
end MProcessor;


architecture Behavorial of MProcessor is

    signal ipb_to_slaves           : ipb_wbus_array(N_SLAVES - 1 downto 0);
    signal ipb_from_slaves         : ipb_rbus_array(N_SLAVES - 1 downto 0);

    signal packet_start            : std_logic                   := '0';
    signal readout_reset           : std_logic                   := '0';
    signal aggregated_stubs        : ldata(7 downto 0)           := (others => LWORD_NULL);
    signal link_aggregator_input   : ldata(N_LINKS - 1 downto 0) := (others => LWORD_NULL);

    signal status_registers        : ipb_reg_v(2 - 1 downto 0)   := (others => (others => '0'));
    signal control_registers       : ipb_reg_v(1 - 1 downto 0)   := (others => (others => '0'));

    signal super_id                : std_logic_vector(31 downto 0) := (others => '0');

begin

    --==============================--
    -- IPBus fabric
    --==============================--


    --==============================--
    fabric: entity work.ipbus_fabric_sel
    --==============================--
    generic map(
        NSLV            => N_SLAVES,
        SEL_WIDTH       => IPBUS_SEL_WIDTH
    )
    port map(
        ipb_in          => ipb_in,
        ipb_out         => ipb_out,
        sel             => ipbus_sel_mprocessor(ipb_in.ipb_addr),
        ipb_to_slaves   => ipb_to_slaves,
        ipb_from_slaves => ipb_from_slaves
    );

    --==============================--
    MProcessorControlInstance: entity work.ipbus_ctrlreg_v
    --==============================--
    generic map(
        N_CTRL    => 1,
        N_STAT    => 2
    )
    port map(
        clk       => ipb_clk,
        reset     => ipb_rst,
        ipbus_in  => ipb_to_slaves(N_SLV_CSR),
        ipbus_out => ipb_from_slaves(N_SLV_CSR),
        d         => status_registers,
        q         => control_registers
    );

    --==============================--
    LinkAggregatorInstance: entity work.LinkAggregatorIPBus
    --==============================--
    generic map (
        N_LINKS       => N_LINKS
    )
    port map (
        --- Input Ports ---
        clk_p         => clk_p,
        packet_start  => packet_start,
        links_in      => link_aggregator_input,
        --- Output Ports ---
        links_out     => aggregated_stubs,
        readout_reset => readout_reset,
        --- IPBus Ports ---
        ipb_clk       => ipb_clk,
        ipb_rst       => ipb_rst,
        ipb_in        => ipb_to_slaves(N_SLV_LINK_AGGREGATOR),
        ipb_out       => ipb_from_slaves(N_SLV_LINK_AGGREGATOR)
    );

    --==============================--
    LinkCombinerInstance : entity work.LinkCombinerIPBus
    --==============================--
    generic map (
        N_INPUT_LINKS => 8
    )
    port map (
        --- Input Ports ---
        clk_p         => clk_p,
        reset         => readout_reset,
        links_in      => aggregated_stubs,
        packet_start  => packet_start,
        header_in     => header_in,
        lff           => gbe_backpressure,
        output_srst   => srst,
        --- Output Ports ---
        link_out      => link_out,
        --- IPBus Ports ---
        ipb_clk       => ipb_clk,
        ipb_rst       => ipb_rst,
        ipb_in        => ipb_to_slaves(N_SLV_LINK_COMBINER),
        ipb_out       => ipb_from_slaves(N_SLV_LINK_COMBINER),
        --- Debug Ports ---
        super_id      => super_id
    );

    -- --==============================--
    -- TrackReconstructorInstance: entity work.TrackReconstructor
    -- --==============================--
    -- port map(
    --     clk_p => clk_p,
    --     reset => readout_reset,
    --     stubs_in => aggregated_stubs(0)
    -- );

    link_aggregator_input(N_LINKS - 1 downto 0) <= links_in;

    status_registers(0)(0) <= gbe_backpressure;
    status_registers(1)    <= super_id;
    packet_start           <= header_start_array(to_integer(unsigned(control_registers(0)(2 downto 0))))(0);

end Behavorial;
