library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
---
use work.ipbus.all;
use work.ipbus_reg_types.all;
use work.ipbus_decode_emp_payload.all;
---
use work.emp_data_types.all;
use work.emp_project_decl.all;
use work.emp_device_decl.all;
use work.emp_ttc_decl.all;
use work.emp_slink_types.all;
---
use work.dtc_link_maps.all;
use work.front_end_data_types.all;
use work.mprocessor_constants.all;
use work.dtc_constants.all;
use work.dtc_data_types.all;


entity emp_payload is
    port (
        --- Input Ports ---
        clk_p            : in  std_logic;
        clk40            : in  std_logic := '0';
        clk_payload      : in  std_logic_vector(2 downto 0);
        rst_payload      : in  std_logic_vector(2 downto 0);
        rst_loc          : in  std_logic_vector(N_REGION - 1 downto 0);
        clken_loc        : in  std_logic_vector(N_REGION - 1 downto 0);
        ctrs             : in  ttc_stuff_array;
        d                : in  ldata(4 * N_REGION - 1 downto 0);
        backpressure     : in  std_logic_vector(SLINK_MAX_QUADS-1 downto 0);
        ttc_cmd          : in  ttc_cmd_t;
        --- Output Ports ---
        bc0              : out std_logic;
        gpio             : out std_logic_vector(29 downto 0);
        gpio_en          : out std_logic_vector(29 downto 0);
        q                : out ldata(4 * N_REGION - 1 downto 0);
        slink_q          : out slink_input_data_quad_array(SLINK_MAX_QUADS-1 downto 0);
        --- IPBus Ports ---
        clk              : in  std_logic;
        rst              : in  std_logic;
        ipb_in           : in  ipb_wbus;
        ipb_out          : out ipb_rbus;
        -- KH
        gbe_q            : out ldata := (others => LWORD_NULL);
        gbe_backpressure : in  std_logic_vector(3 downto 0)
    );
end entity emp_payload;


architecture rtl of emp_payload is


    -- IPBus fabric

    signal ipb_to_slaves                : ipb_wbus_array(N_SLAVES - 1 downto 0);
    signal ipb_from_slaves              : ipb_rbus_array(N_SLAVES - 1 downto 0);

    signal fe_control_registers         : ipb_reg_v(0 downto 0);
    signal fe_status_registers          : ipb_reg_v(0 downto 0);
    signal mproc_channel_sel            : ipb_reg_v(0 downto 0);
    signal link_aggr_control            : ipb_reg_v(0 downto 0);


    -- FE data extraction and monitoring

    signal ipb_chain                    : ipbdc_bus_array(cNumberOfFEModules downto 0);
    signal stubs                        : ldata(cNumberOfFEModules - 1 downto 0);
    signal header_array                 : tCICHeaderArray(cNumberOfFEModules*cNumberOfCICs - 1 downto 0) := (others => ('0', (others => '0'), (others => '0'), (others => '0')));
    signal header_start_array           : std_logic_vector(cNumberOfFEModules*cNumberOfCICs - 1 downto 0) := (others => '0');
    
    signal eth_link_out                 : ldata(3 downto 0)                        := (others => LWORD_NULL);

    

    -- MProcessor
    constant cNumberOfMProcessors                  : integer                                  := 2;
    signal mprocessor_ipb_chain                    : ipbdc_bus_array(cNumberOfMProcessors downto 0);
    signal srst                                    : std_logic                                := '0';
    signal ttc_resync_reset, ttc_resync_reset_prev : std_logic                                := '0';
    signal ttc_bc0, ttc_oc0, ttc_resync            : std_logic                                := '0';

    signal orbit_counter                : unsigned(31 downto 0)                    := to_unsigned(0   , 32);
    signal bunch_counter                : unsigned(11 downto 0)                    := to_unsigned(3564, 12);


    -- Daqpath

    signal daq_read                   : tDaqFlagLinkArray(cNumberOfFEModules - 1 downto 0);
    signal daq_data                   : tDaqDataLinkArray(cNumberOfFEModules - 1 downto 0);
    signal daq_empty                  : tDaqFlagLinkArray(cNumberOfFEModules - 1 downto 0);


    -- TCDS & Fast commands
    
    signal global_fcmd                : tFastCommand;
    signal tcds_fcmd                  : tFastCommand;


begin

  
    --==============================--
    -- IPBus fabric
    --==============================--

    --==============================--
    fabric: entity work.ipbus_fabric_sel
    --==============================--
        generic map (
            NSLV      => N_SLAVES,
            SEL_WIDTH => IPBUS_SEL_WIDTH
        )
        port map (
            ipb_in          => ipb_in,
            ipb_out         => ipb_out,
            sel             => ipbus_sel_emp_payload(ipb_in.ipb_addr),
            ipb_to_slaves   => ipb_to_slaves,
            ipb_from_slaves => ipb_from_slaves
        );

    --==============================--
    fe_channel_ctrl : entity work.ipbus_ctrlreg_v
    --==============================--
    generic map(
        N_CTRL    => 1,
        N_STAT    => 1
    )
    port map (
        clk       => clk,
        reset     => rst,
        ipbus_in  => ipb_to_slaves(N_SLV_FE),
        ipbus_out => ipb_from_slaves(N_SLV_FE),
        d         => fe_status_registers,
        q		  => fe_control_registers
    );

    fe_status_registers(0) <= std_logic_vector(to_unsigned(cNumberOfFEModules, 32));

    --==============================--
    channel_select : entity work.ipbus_dc_fabric_sel
    --==============================--
    generic map (
        SEL_WIDTH => 7
    )
    port map (
        clk       => clk,
        rst       => rst,
        sel       => fe_control_registers(0)(6 downto 0),
        ipb_in    => ipb_to_slaves(N_SLV_FE_CHAN),
        ipb_out   => ipb_from_slaves(N_SLV_FE_CHAN),
        ipbdc_out => ipb_chain(0),
        ipbdc_in  => ipb_chain(cNumberOfFEModules)
    );

    --==============================--
    mproc_channel_ctrl: entity work.ipbus_reg_v
    --==============================--
    generic map(
        N_REG           => 1
    )
    port map(
        clk             => clk,
        reset           => rst,
        ipbus_in        => ipb_to_slaves(N_SLV_MPROC_CTRL),
        ipbus_out       => ipb_from_slaves(N_SLV_MPROC_CTRL),
        q               => mproc_channel_sel,
        qmask           => (0 => X"00000fff")
    );


    


    --==============================--
    -- FE data extraction and monitoring
    --==============================--

    --==============================--
    genLinkInterface: for i in 0 to cNumberOfFEModules - 1 generate
    --==============================--

        signal ipb_to_channel   : ipb_wbus;
        signal ipb_from_channel : ipb_rbus;

    begin

        --==============================--
        channel_node : entity work.ipbus_dc_node
        --==============================--
        generic map (
            I_SLV     => i,
            SEL_WIDTH => 7,
            PIPELINE  => false
        )
        port map (
            clk       => clk,
            rst       => rst,
            ipb_out   => ipb_to_channel,
            ipb_in    => ipb_from_channel,
            ipbdc_in  => ipb_chain(i),
            ipbdc_out => ipb_chain(i + 1)
        );

        --==============================--
        LinkInterfaceInstance : entity work.LinkInterface
        --==============================--
        generic map (
            module_type => cDTCInputLinkMap(i).module_type,
            data_rate   => cDTCInputLinkMap(i).data_rate,
            cic_type    => cDTCInputLinkMap(i).cic_type
        )
        port map (
            --- Input Ports ---
            clk_p               => clk_p,
            clk40               => clk40,
            link_in             => d(cDTCInputLinkMap(i).index),
            daq_read            => daq_read(i),
            global_fcmd         => global_fcmd,
            --- Output Ports ---
            link_out            => q(cDTCInputLinkMap(i).index),
            stub_out            => stubs(i),
            header_out          => header_array(cNumberOfCICs * i + cNumberOfCICs - 1 downto cNumberOfCICs * i),
            daq_out             => daq_data(i),
            daq_empty           => daq_empty(i),
            --- IPBus Ports ---
            clk                 => clk,
            rst                 => rst,
            ipb_in              => ipb_to_channel,
            ipb_out             => ipb_from_channel,
            --- Debug Ports ---
            debug_header_start  => header_start_array(cNumberOfCICs*(i + 1) - 1 downto cNumberOfCICs*i)
        );
    end generate genLinkInterface;


    --==============================--
    -- Global TCDS
    --==============================--

    --==============================--
    GlobalFastCommand: entity work.GlobalFastCommand
    --==============================--
    port map (
        --- Input Ports ---
        clk_p       => clk_p,
        clk40       => clk40,
        ext_fcmd    => tcds_fcmd,
        --- Output Ports ---
        global_fcmd => global_fcmd,
        --- IPBus Ports ---
        clk         => clk,                       
        rst         => rst,                       
        ipb_in      => ipb_to_slaves(N_SLV_TCDS_FAST_CMD),  
        ipb_out     => ipb_from_slaves(N_SLV_TCDS_FAST_CMD)
    );

    tcds_fcmd.fast_reset    <= ttc_resync;
    tcds_fcmd.l1a_trig      <= '0';
    tcds_fcmd.cal_pulse     <= '0';
    tcds_fcmd.counter_reset <= ttc_bc0;





    --==============================--
    -- L1 Data Aggregator
    --==============================--

    --==============================--
    L1DataAggregator : entity work.L1DataAggregator
    --==============================--
    port map (
        --- Input Ports ---
        clk_p   => clk_p,
        daq_in  => daq_data,
        empty   => daq_empty,
        --- Output Ports ---
        read    => daq_read,
        --- IPBus Ports ---
        clk     => clk,
        rst     => rst,
        ipb_in  => ipb_to_slaves(N_SLV_BE_DAQ),
        ipb_out => ipb_from_slaves(N_SLV_BE_DAQ)
    );

    --==============================--
    pRouteStubsToOutput : process (clk_p) is
    --==============================--
    begin
        if rising_edge(clk_p) then
            for i in 0 to cNumberOfOutputLinks - 1 loop
                q(cDTCOutputLinkMap(i)).valid  <= d(cDTCInputLinkMap(i).index).valid;
                q(cDTCOutputLinkMap(i)).data   <= d(cDTCInputLinkMap(i).index).data;
                q(cDTCOutputLinkMap(i)).strobe <= '1';
            end loop;
        end if;
    end process pRouteStubsToOutput;


    
    --==============================--
    LinkMonitorInterfaceInstance : entity work.LinkMonitorInterface
    --==============================--
    generic map (
        N_LINKS => cNumberOfFEModules
    )
    port map (
        --- Input Ports ---
        clk_p => clk_p,
        stubs => stubs(cNumberOfFEModules - 1 downto 0),
        headers => header_array(2*cNumberOfFEModules - 1 downto 0),
        header_start => header_start_array(2*cNumberOfFEModules - 1 downto 0),
        --- IPBus Ports ---
        clk     => clk,
        rst     => rst,
        ipb_in  => ipb_to_slaves(N_SLV_LINK_MONITOR),
        ipb_out => ipb_from_slaves(N_SLV_LINK_MONITOR)
    );



    --==============================--
    -- Link Aggregation and Combination
    --==============================--

    --==============================--
    mprocessor_channel_select: entity work.ipbus_dc_fabric_sel
    --==============================--
    generic map(
        SEL_WIDTH       => 7
    )
    port map(
        clk             => clk,
        rst             => rst,
        sel             => mproc_channel_sel(0)(6 downto 0),
        ipb_in          => ipb_to_slaves(N_SLV_MPROCESSOR),
        ipb_out         => ipb_from_slaves(N_SLV_MPROCESSOR),
        ipbdc_out       => mprocessor_ipb_chain(0),
        ipbdc_in        => mprocessor_ipb_chain(cNumberOfMProcessors)
    );


    --==============================-
    genMProcessors : for i in 0 to cNumberOfMProcessors - 1 generate
    --==============================-

        signal ipb_to_channel        : ipb_wbus;
        signal ipb_from_channel      : ipb_rbus;
        constant cNumberofInputLinks : integer := cNumberOfFEModules;

    begin

        --==============================--
        channel_node: entity work.ipbus_dc_node
        --==============================--
        generic map(
            I_SLV       => i,
            SEL_WIDTH   => 7,
            PIPELINE    => false
        )
        port map(
            clk         => clk,
            rst         => rst,
            ipb_out     => ipb_to_channel,
            ipb_in      => ipb_from_channel,
            ipbdc_in    => mprocessor_ipb_chain(i),
            ipbdc_out   => mprocessor_ipb_chain(i + 1)
        );

        --==============================--
        MProcessorInstance: entity work.MProcessor
        --==============================--
        generic map(
            N_LINKS => cNumberofInputLinks
        )
        port map(
            --- Input Ports ---
            clk_p              => clk_p,
            links_in           => stubs(cNumberofInputLinks - 1 downto 0),
            header_in          => header_array,
            header_start_array => convertSLVtoHeaderStartArray(header_start_array),
            gbe_backpressure   => gbe_backpressure(i),
            srst               => srst,
            --- Output Ports ---
            link_out           => eth_link_out(i),
            --- IPBus Ports ---
            ipb_clk            => clk,
            ipb_rst            => rst,
            ipb_in             => ipb_to_channel,
            ipb_out            => ipb_from_channel
        );
    end generate genMProcessors;


    gbe_q <= eth_link_out;

    
    --==============================--
    pTTC: process(clk40)
    --==============================--
    begin
        if rising_edge(clk40) then

            ttc_bc0    <= '0';
            ttc_oc0    <= '0';
            ttc_resync <= '0';

            if ttc_cmd = TTC_BCMD_BC0 then
                orbit_counter <= orbit_counter + 1;
                bunch_counter <= (others => '0');
                ttc_bc0 <= '1';

                if ttc_resync_reset = '1' and ttc_resync_reset_prev = '0' then

                    ttc_resync <= '1';

                end if;

                ttc_resync_reset_prev <= ttc_resync_reset;


            elsif ttc_cmd = TTC_BCMD_OC0 then

                orbit_counter <= (others => '0');
                bunch_counter <= to_unsigned(3564, bunch_counter'length);
                ttc_oc0 <= '1';

            elsif bunch_counter < 3564 then

                bunch_counter <= bunch_counter + 1;     
                   
            end if;

        end if;
    end process;

    ttc_resync_reset <= mproc_channel_sel(0)(8);
    srst <= mproc_channel_sel(0)(7) or ttc_oc0;

    q(20).valid  <= '1';
    q(20).data   <= "0000" & std_logic_vector(bunch_counter) & "0000000000000000" & std_logic_vector(orbit_counter);
    q(20).start  <= '1';
    q(20).strobe <= '1';




    bc0     <= '0';
    gpio    <= (others => '0');
    gpio_en <= (others => '0');


    --==============================--
    slink_q_gen : for q in SLINK_MAX_QUADS-1 downto 0 generate
    --==============================--
    begin
        slink_q(q) <= SLINK_INPUT_DATA_ARRAY_NULL;
    end generate slink_q_gen;



end architecture rtl;
